// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/dev_fifo.c $
// $Rev: 30377 $
// $Date: 2014-09-18 12:09:04 -0500 (Thu, 18 Sep 2014) $

#include "dev_fifo.h"

#include "gsc_sleep.h"
#include "gsc_version.h"

#include "dev_config.h"
#include "dev_reg.h"



/******************************************************************************
*
*	Function:	_almost_level_fifo
*
*	Purpose:
*
*		Implement the FIFO Almost Level parameter service for a single Tx/Rx
*		Almost Full/Almost Empty level. Validate and apply the setting, as
*		appropriate, then retrieve the current setting for the caller, also
*		as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		The specific level to which to apply the effort.
*
*		reg			The FIFO register to access.
*
*		full		Access the Almost Full level?
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _almost_level_fifo(
	dev_data_t*	dev,
	U32			which,
	U32			reg,
	U32			full,
	UL			set,
	UL*			get)
{
	U32	ae;
	U32	af;
	U32	almost;
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= GSC_REG_READ(dev, reg, &almost);
		if (status != (U32) GSC_SUCCESS)	break;

		af	= HPDI32_DECODE_FULL(almost);
		ae	= HPDI32_DECODE_EMPTY(almost);

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			if (which & HPDI32_WHICH_AF)
				set	= HPDI32_FIFO_ALMOST_FULL_DEFAULT;
			else
				set	= HPDI32_FIFO_ALMOST_EMPTY_DEFAULT;
		}
		else if (set <= HPDI32_FIFO_ALMOST_LEVEL_MAX)
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else
		{
			set	= HPDI32_FIFO_ALMOST_LEVEL_MAX;
		}

		if (set != GSC_NO_CHANGE)
		{
			if (which & HPDI32_WHICH_TX)
			{
				if ((dev->cache.tfsr) && (set > dev->cache.tfsr))
					set	= dev->cache.tfsr;
			}
			else
			{
				if ((dev->cache.rfsr) && (set > dev->cache.rfsr))
					set	= dev->cache.rfsr;
			}

			if (full)
				almost	= HPDI32_ALMOST_ENCODE(set, ae);
			else
				almost	= HPDI32_ALMOST_ENCODE(af, set);

			status	= GSC_REG_WRITE(dev, reg, almost);
			if (status != (U32) GSC_SUCCESS)	break;

			if (which & HPDI32_WHICH_TX)
				status	= GSC_REG_MOD(dev, HPDI32_BCR, (U32) ~0UL, HPDI32_BCR_TX_FIFO_RESET);
			else
				status	= GSC_REG_MOD(dev, HPDI32_BCR, (U32) ~0UL, HPDI32_BCR_RX_FIFO_RESET);

			gsc_sleep_ms(1);	// Sleep for 1ms though a FIFO reset takes less than 10us.
		}

		if (get)
		{
			status	= GSC_REG_READ(dev, reg, &almost);
			if (status != (U32) GSC_SUCCESS)	break;

			if (full)
				get[0]	= HPDI32_DECODE_FULL(almost);
			else
				get[0]	= HPDI32_DECODE_EMPTY(almost);
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	fifo_almost_level
*
*	Purpose:
*
*		Implement the FIFO Almost Level parameter service. Validate and apply
*		the setting(s), as appropriate, then retrieve the current setting(s)
*		for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 fifo_almost_level(dev_data_t* dev, U32 which, UL set, UL* get)
{
	static const struct
	{
		U32	which;
		U32	reg;
		U32	full;
	} list[]	=
	{
		// which								reg			full
		{ HPDI32_WHICH_TX | HPDI32_WHICH_AF,	HPDI32_TAR,	1		},
		{ HPDI32_WHICH_TX | HPDI32_WHICH_AE,	HPDI32_TAR,	0		},
		{ HPDI32_WHICH_RX | HPDI32_WHICH_AF,	HPDI32_RAR,	1		},
		{ HPDI32_WHICH_RX | HPDI32_WHICH_AE,	HPDI32_RAR,	0		}
	};

	U32	bcr;
	int	i;
	U32	stat;
	U32	status	= (U32) GSC_SUCCESS;

	status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX_AF_AE);

	if (status == (U32) GSC_SUCCESS)
	{
		for (i = 0; i < (int) ARRAY_ELEMENTS(list); i++)
		{
			if ((list[i].which & which) != list[i].which)
				continue;

			stat	= GSC_REG_READ(dev, HPDI32_BCR, &bcr);
			status	= (status == (U32) GSC_SUCCESS) ? stat : status;
			if (stat != (U32) GSC_SUCCESS)	continue;

			stat	= _almost_level_fifo(	dev,
											list[i].which,
											list[i].reg,
											list[i].full,
											set,
											get);
			status	= (status == (U32) GSC_SUCCESS) ? stat : status;
			if (stat != (U32) GSC_SUCCESS)	continue;
		}
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reset_fifo
*
*	Purpose:
*
*		Implement the FIFO Reset parameter service for the given FIFO.
*		Validate and apply the setting(s), as appropriate, then retrieve the
*		current setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		bcr			The BCR bit for the FIFO to reset.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _reset_fifo(dev_data_t* dev, U32 bcr, UL set, UL* get)
{
	U32	reset	= HPDI32_FIFO_RESET_NO;
	U32	status	= (U32) GSC_SUCCESS;

	for (;;)	// A convenience loop.
	{
		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_FIFO_RESET_DEFAULT;
		}
		else if (	(set == HPDI32_FIFO_RESET_YES) ||
					(set == HPDI32_FIFO_RESET_NO))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_FIFO_RESET_YES;
		}
		else
		{
			set	= HPDI32_FIFO_RESET_NO;
		}

		if ((set != GSC_NO_CHANGE) && (set != HPDI32_FIFO_RESET_NO))
		{
			reset	= HPDI32_FIFO_RESET_YES;
			status	= GSC_REG_MOD(dev, HPDI32_BCR, bcr, bcr);
			gsc_sleep_ms(1);	// Sleep for 1ms though a FIFO reset takes less than 10us.
			if (status != (U32) GSC_SUCCESS)	break;
		}

		if (get)
			get[0]	= reset;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	fifo_reset
*
*	Purpose:
*
*		Implement the FIFO Reset parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 fifo_reset(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	stat;
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX);
		if (status != (U32) GSC_SUCCESS)	break;

		for (; which & HPDI32_WHICH_TX;)
		{
			stat	= _reset_fifo(dev, HPDI32_BCR_TX_FIFO_RESET, set, get);
			status	= (status == (U32) GSC_SUCCESS) ? stat : status;
			break;
		}

		for (; which & HPDI32_WHICH_RX;)
		{
			stat	= _reset_fifo(dev, HPDI32_BCR_RX_FIFO_RESET, set, get);
			status	= (status == (U32) GSC_SUCCESS) ? stat : status;
			break;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	fifo_size
*
*	Purpose:
*
*		Implement the FIFO Size parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 fifo_size(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET is not supported for this parameter.

		if (which & HPDI32_WHICH_TX)
		{
			if (get)
				get[0]	= dev->cache.tfsr;
		}

		if (which & HPDI32_WHICH_RX)
		{
			if (get)
				get[0]	= dev->cache.rfsr;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	fifo_status
*
*	Purpose:
*
*		Implement the FIFO Status parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 fifo_status(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	bsr;
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX);
		if (status != (U32) GSC_SUCCESS)	break;

		status	= GSC_REG_READ(dev, HPDI32_BSR, &bsr);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET is not supported for this parameter.

		if (get)
		{
			if (which & HPDI32_WHICH_TX)
			{
				if ((bsr & HPDI32_BSR_TX_NOT_EMPTY) == 0)
					get[0]	= HPDI32_FIFO_STATUS_EMPTY;
				else if ((bsr & HPDI32_BSR_TX_NOT_AE) == 0)
					get[0]	= HPDI32_FIFO_STATUS_ALMOST_EMPTY;
				else if ((bsr & HPDI32_BSR_TX_NOT_FULL) == 0)
					get[0]	= HPDI32_FIFO_STATUS_FULL;
				else if ((bsr & HPDI32_BSR_TX_NOT_AF) == 0)
					get[0]	= HPDI32_FIFO_STATUS_ALMOST_FULL;
				else
					get[0]	= HPDI32_FIFO_STATUS_MEDIAN;
			}

			if (which & HPDI32_WHICH_RX)
			{
				if ((bsr & HPDI32_BSR_RX_NOT_EMPTY) == 0)
					get[0]	= HPDI32_FIFO_STATUS_EMPTY;
				else if ((bsr & HPDI32_BSR_RX_NOT_AE) == 0)
					get[0]	= HPDI32_FIFO_STATUS_ALMOST_EMPTY;
				else if ((bsr & HPDI32_BSR_RX_NOT_FULL) == 0)
					get[0]	= HPDI32_FIFO_STATUS_FULL;
				else if ((bsr & HPDI32_BSR_RX_NOT_AF) == 0)
					get[0]	= HPDI32_FIFO_STATUS_ALMOST_FULL;
				else
					get[0]	= HPDI32_FIFO_STATUS_MEDIAN;
			}
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	fifo_transfer_size
*
*	Purpose:
*
*		Implement the FIFO Transfer Size parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for  the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 fifo_transfer_size(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET is not supported for this parameter.

		if (get)
		{
			if (which & HPDI32_WHICH_TX)
			{
				status	= io_pio_write_available(dev, get);
				if (status != (U32) GSC_SUCCESS)	break;
			}

			if (which & HPDI32_WHICH_RX)
			{
				status	= io_pio_read_available(dev, get);
				if (status != (U32) GSC_SUCCESS)	break;
			}
		}

		break;
	}

	return(status);
}


