// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/dev_io.c $
// $Rev: 30377 $
// $Date: 2014-09-18 12:09:04 -0500 (Thu, 18 Sep 2014) $

#include "dev_io.h"

#include "dev_access.h"
#include "dev_config.h"
#include "dev_reg.h"



/******************************************************************************
*
*	Function:	_io_rx_funcs_select
*
*	Purpose:
*
*		Perform a check to see which I/O functions will be needed when a read
*		is requested.
*
*	Arguments:
*
*		dev			The device to access.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _io_rx_funcs_select(dev_data_t* dev)
{
	dev->rx.cache.pio_funcs		= &io_pio_read_funcs;

	if (dev->rx.gio.dma_enable == 0)
	{
		// PIO
		dev->rx.cache.dma_need		= 0;
		dev->rx.cache.mode_funcs	= &io_pio_read_funcs;
	}
	else if (dev->rx.gio.dma_demand)
	{
		// Demand Mode DMA
		dev->rx.cache.dma_need	= GSC_DMA_FLAG_DMDMA_RX;

		if (dev->cache.single_cycle_disable)
			dev->rx.cache.mode_funcs	= &io_dmdma_read_scd_funcs;
		else if (dev->rx.single_cycle == HPDI32_IO_SINGLE_CYCLE_PRESENT)
			dev->rx.cache.mode_funcs	= &io_dmdma_read_scp_funcs;
		else
			dev->rx.cache.mode_funcs	= &io_dmdma_read_sca_funcs;
	}
	else
	{
		// Non Demand Mode DMA
		dev->rx.cache.dma_need	= GSC_DMA_FLAG_BLDMA_RX;

		if (dev->rx.gio.dma_control_mode == HPDI32_IO_DMA_CONTROL_MODE_AUTOMATIC)
			dev->rx.cache.mode_funcs	= &io_dma_read_auto_funcs;
		else
			dev->rx.cache.mode_funcs	= &io_dma_read_manual_funcs;
	}
}



/******************************************************************************
*
*	Function:	_io_common_parameter_which
*
*	Purpose:
*
*		Provide common code for processing I/O parameters based on the WHICH
*		argument.
*
*	Arguments:
*
*		dev			The device to access.
*
*		func		The function to call for parameter specific processing.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_common_parameter_which(
	dev_data_t*	dev,
	U32			(*func)(dev_data_t*	dev,
						dev_io_t*	io,
						UL			set,
						UL*			get),
	U32			which,
	UL			set,
	UL*			get)
{
	U32	stat;
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX);
		if (status != (U32) GSC_SUCCESS)	break;

		for (; which & HPDI32_WHICH_TX;)
		{
			stat	= (func)(dev, &dev->tx, set, get);
			status	= (status == (U32) GSC_SUCCESS) ? stat : status;
			break;
		}

		for (; which & HPDI32_WHICH_RX;)
		{
			stat	= (func)(dev, &dev->rx, set, get);
			status	= (status == (U32) GSC_SUCCESS) ? stat : status;
			break;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_io_abort_side
*
*	Purpose:
*
*		Implement the I/O Abort parameter service for a single Tx/Rx side.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_abort_side(dev_data_t* dev, dev_io_t* io, UL set, UL* get)
{
	U32	status	= (U32) GSC_SUCCESS;

	for (;;)	// A convenience loop.
	{
		// Ordinarily we would get exclusive access. We don't do that
		// here since an in progress I/O already has exclusive access.

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_ABORT_DEFAULT;
		}
		else if (	(set == HPDI32_IO_ABORT_NO) ||
					(set == HPDI32_IO_ABORT_YES))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_IO_ABORT_YES;
		}
		else
		{
			set	= HPDI32_IO_ABORT_NO;
		}

		if (set == HPDI32_IO_ABORT_YES)
			status	= gsc_io_abort(&io->gio);
		else if (set != GSC_NO_CHANGE)
			io->gio.abort	= set;

		if (get)
			get[0]	= io->gio.abort;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	io_abort
*
*	Purpose:
*
*		Implement the I/O Abort parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_abort(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_abort_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_aborted_side
*
*	Purpose:
*
*		Implement the I/O Aborted parameter service for a single Tx/Rx side.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_aborted_side(dev_data_t* dev, dev_io_t* io, UL set, UL* get)
{
	U32	status	= (U32) GSC_SUCCESS;

	for (;;)	// A convenience loop.
	{
		// Ordinarily we would get exclusive access. We don't do that
		// here since an in progress I/O already has exclusive access.

		// SET is not supported by this parameter.

		if (get)
			get[0]	= io->gio.aborted;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	io_aborted
*
*	Purpose:
*
*		Implement the I/O Aborted parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_aborted(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_aborted_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_callback_arg_side
*
*	Purpose:
*
*		Implement the I/O Callback Argument parameter service for a single
*		Tx/Rx side. Validate and apply the setting, as appropriate, then
*		retrieve the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_callback_arg_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32			status	= (U32) GSC_FAILED;
	gsc_event_t	tmp;

	for (;;)	// A convenience loop.
	{
		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_CALLBACK_ARG_DEFAULT;
		}

		if (set != GSC_NO_CHANGE)
		{
			status	= gsc_io_callback_config(
						&io->gio,
						(void*) GSC_NO_CHANGE,	// id: index
						(void*) GSC_NO_CHANGE,	// function
						dev,					// arg1: device handle
						GSC_NO_CHANGE,			// arg2: I/O status
						set,					// arg3: user argument
						NULL);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		if (get)
		{
			status	= gsc_io_callback_config(
						&io->gio,
						(void*) GSC_NO_CHANGE,		// id: index
						(void*) GSC_NO_CHANGE,		// function
						(void*) GSC_NO_CHANGE,		// arg1: device_handle
						GSC_NO_CHANGE,				// arg2: I/O status
						GSC_NO_CHANGE,				// arg3: user argument
						&tmp);
			if (status != (U32) GSC_SUCCESS)	break;
			get[0]	= tmp.arg3;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	io_callback_arg
*
*	Purpose:
*
*		Implement the I/O Callback Argument parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_callback_arg(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_callback_arg_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_callback_func_side
*
*	Purpose:
*
*		Implement the I/O Callback Function parameter service for a single
*		Tx/Rx side. Validate and apply the setting, as appropriate, then
*		retrieve the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_callback_func_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32			status	= (U32) GSC_FAILED;
	gsc_event_t	tmp;

	for (;;)	// A convenience loop.
	{
		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_CALLBACK_FUNC_DEFAULT;
		}

		if (set != GSC_NO_CHANGE)
		{
			status	= gsc_io_callback_config(
						&io->gio,
						(void*) GSC_NO_CHANGE,	// id: index
						(void*) set,			// function
						dev,					// arg1: device handle
						GSC_NO_CHANGE,			// arg2: I/O status
						GSC_NO_CHANGE,			// arg3: user argument
						NULL);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		if (get)
		{
			status	= gsc_io_callback_config(
						&io->gio,
						(void*) GSC_NO_CHANGE,	// id: index
						(void*) GSC_NO_CHANGE,	// function
						(void*) GSC_NO_CHANGE,	// arg1: device handle
						GSC_NO_CHANGE,			// arg2: I/O status
						GSC_NO_CHANGE,			// arg3: user argument
						&tmp);
			if (status != (U32) GSC_SUCCESS)	break;
			get[0]	= (UL) tmp.callback;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	io_callback_func
*
*	Purpose:
*
*		Implement the I/O Callback Function parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_callback_func(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_callback_func_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_data_size_side
*
*	Purpose:
*
*		Implement the I/O Data Size parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_data_size_side(dev_data_t* dev, dev_io_t* io, UL set, UL* get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_DATA_SIZE_DEFAULT;
		}
		else if (	(set == HPDI32_IO_DATA_SIZE_8_BITS) ||
					(set == HPDI32_IO_DATA_SIZE_16_BITS) ||
					(set == HPDI32_IO_DATA_SIZE_32_BITS))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set > HPDI32_IO_DATA_SIZE_16_BITS)
		{
			set	= HPDI32_IO_DATA_SIZE_32_BITS;
		}
		else if (set > HPDI32_IO_DATA_SIZE_8_BITS)
		{
			set	= HPDI32_IO_DATA_SIZE_16_BITS;
		}
		else
		{
			set	= HPDI32_IO_DATA_SIZE_8_BITS;
		}

		if ((set != GSC_NO_CHANGE) && (set != io->gio.data_size))
		{
			io->gio.data_size	= set;
			gsc_dma_channel_close(&dev->dma, &io->gio.dma_channel);
		}

		if (get)
			get[0]	= io->gio.data_size;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_data_size
*
*	Purpose:
*
*		Implement the I/O Data Size parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_data_size(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_data_size_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_dma_channel_select_side
*
*	Purpose:
*
*		Implement the I/O DMA Channel Select parameter service for a single
*		Tx/Rx side. Validate and apply the setting, as appropriate, then
*		retrieve the current setting for the caller, also as appropriate.
*		Since this parameter affect the DMA Channel Selection process, the
*		DMA Channel Selection selection is evaluated.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_dma_channel_select_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			if ((io->gio.is_tx) && (dev->favor_tx))
				set	= HPDI32_IO_DMA_CHANNEL_SEL_TX_DEFAULT;
			else
				set	= HPDI32_IO_DMA_CHANNEL_SEL_RX_DEFAULT;
		}
		else if (	(set == HPDI32_IO_DMA_CHANNEL_SEL_STATIC) ||
					(set == HPDI32_IO_DMA_CHANNEL_SEL_DYNAMIC))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_IO_DMA_CHANNEL_SEL_STATIC;
		}
		else
		{
			set	= HPDI32_IO_DMA_CHANNEL_SEL_DYNAMIC;
		}

		if (set != GSC_NO_CHANGE)
		{
			io->gio.dma_dynamic	= set;

			if (set == HPDI32_IO_DMA_CHANNEL_SEL_DYNAMIC)
				gsc_dma_channel_close(&dev->dma, &io->gio.dma_channel);
		}

		if (get)
			get[0]	= io->gio.dma_dynamic;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_dma_channel_select
*
*	Purpose:
*
*		Implement the I/O DMA Channel Select parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_dma_channel_select(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_dma_channel_select_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_dma_control_mode_side
*
*	Purpose:
*
*		Implement the I/O DMA Control Mode parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate. Since this
*		parameter affect the DMA Channel Selection process, the DMA Channel
*		Selection selection is evaluated.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_dma_control_mode_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_DMA_CONTROL_MODE_DEFAULT;
		}
		else if (	(set == HPDI32_IO_DMA_CONTROL_MODE_AUTOMATIC) ||
					(set == HPDI32_IO_DMA_CONTROL_MODE_MANUAL))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_IO_DMA_CONTROL_MODE_AUTOMATIC;
		}
		else
		{
			set	= HPDI32_IO_DMA_CONTROL_MODE_MANUAL;
		}

		if (set != GSC_NO_CHANGE)
		{
			io->gio.dma_control_mode	= set;

			if (io->gio.is_tx)
				io_tx_funcs_select(dev);
			else
				_io_rx_funcs_select(dev);
		}

		if (get)
			get[0]	= io->gio.dma_control_mode;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_dma_control_mode
*
*	Purpose:
*
*		Implement the I/O DMA Control Mode parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_dma_control_mode(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_dma_control_mode_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_dma_priority_side
*
*	Purpose:
*
*		Implement the I/O DMA Priority parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_dma_priority_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			if ((io->gio.is_tx) && (dev->favor_tx))
				set	= HPDI32_IO_DMA_PRIORITY_TX_DEFAULT;
			else
				set	= HPDI32_IO_DMA_PRIORITY_RX_DEFAULT;
		}
		else if (	(set == HPDI32_IO_DMA_PRIORITY_ENABLE) ||
					(set == HPDI32_IO_DMA_PRIORITY_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_IO_DMA_PRIORITY_ENABLE;
		}
		else
		{
			set	= HPDI32_IO_DMA_PRIORITY_DISABLE;
		}

		if ((set != GSC_NO_CHANGE) && (set != io->gio.dma_priority))
		{
			io->gio.dma_priority	= set;
			gsc_dma_channel_close(&dev->dma, &io->gio.dma_channel);
		}

		if (get)
			get[0]	= io->gio.dma_priority;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_dma_priority
*
*	Purpose:
*
*		Implement the I/O DMA Priority parameter service. Validate and apply
*		the setting(s), as appropriate, then retrieve the current setting(s)
*		for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_dma_priority(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_dma_priority_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_buffer_pointer_side
*
*	Purpose:
*
*		Implement the I/O Buffer Pointer parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate. This is a
*		GET only service.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_buffer_pointer_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32			status;
	gsc_mem_t	tmp;
	U32			unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET not supported for this parameter.

		if (get)
		{
			status	= gsc_mem_alloc(&io->mem, GSC_NO_CHANGE, &tmp);
			if (status != (U32) GSC_SUCCESS)	break;
			get[0]	= tmp.usr_adrs;
		}

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_buffer_pointer
*
*	Purpose:
*
*		Implement the I/O Buffer Pointer parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate. This is a GET only
*		service.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_buffer_pointer(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_buffer_pointer_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_buffer_size_side
*
*	Purpose:
*
*		Implement the I/O Buffer Size parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_buffer_size_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32			status		= (U32) GSC_SUCCESS;
	gsc_mem_t	tmp;
	U32			unlock_rx	= 0;
	U32			unlock_tx	= 0;

	for (;;)	// A convenience loop.
	{
		// Lock both as either might be in use.
		status		= gsc_sem_lock(&io->gio.dev->rx.gio.sem);
		unlock_rx	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		status		= gsc_sem_lock(&io->gio.dev->tx.gio.sem);
		unlock_tx	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
			;
		else if (set == GSC_DEFAULT)
			set	= HPDI32_IO_BUFFER_SIZE_DEFAULT;

		if (set != GSC_NO_CHANGE)
		{
			status	= gsc_mem_alloc(&io->mem, set, NULL);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		if (get)
		{
			status	= gsc_mem_alloc(&io->mem, GSC_NO_CHANGE, &tmp);
			if (status != (U32) GSC_SUCCESS)	break;
			get[0]	= tmp.size;
		}

		break;
	}

	if (unlock_rx)
		gsc_sem_unlock(&io->gio.dev->rx.gio.sem);

	if (unlock_tx)
		gsc_sem_unlock(&io->gio.dev->tx.gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_buffer_size
*
*	Purpose:
*
*		Implement the I/O Driver Buffer Size parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate. This is a GET only
*		service.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_buffer_size(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_buffer_size_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_mode_side
*
*	Purpose:
*
*		Implement the I/O Mode parameter service for a single Tx/Rx side.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting for the caller, also as appropriate. Since this
*		parameter affect the DMA Channel Selection process, the DMA Channel
*		Selection selection is evaluated.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_mode_side(dev_data_t* dev, dev_io_t* io, UL set, UL* get)
{
	U32	dma_demand;
	U32	dma_enable;
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_MODE_DEFAULT;
		}
		else if (	(set == HPDI32_IO_MODE_PIO) ||
					(set == HPDI32_IO_MODE_DMA) ||
					(set == HPDI32_IO_MODE_DMDMA))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else
		{
			set	= HPDI32_IO_MODE_DEFAULT;
		}

		if (set != GSC_NO_CHANGE)
		{
			if (set == HPDI32_IO_MODE_DMDMA)
			{
				dma_demand	= 1;
				dma_enable	= 1;
			}
			else if (set == HPDI32_IO_MODE_DMA)
			{
				dma_demand	= 0;
				dma_enable	= 1;
			}
			else // set == HPDI32_IO_MODE_PIO
			{
				dma_demand	= 0;
				dma_enable	= 0;
			}

			if ((dma_demand != io->gio.dma_demand) ||
				(dma_enable != io->gio.dma_enable))
			{
				io->gio.dma_demand	= dma_demand;
				io->gio.dma_enable	= dma_enable;
				gsc_dma_channel_close(&dev->dma, &io->gio.dma_channel);

				if (io->gio.is_tx)
					io_tx_funcs_select(dev);
				else
					_io_rx_funcs_select(dev);
			}
		}

		if (get)
		{
			if (io->gio.dma_enable == 0)
				get[0]	= HPDI32_IO_MODE_PIO;
			else if (io->gio.dma_demand)
				get[0]	= HPDI32_IO_MODE_DMDMA;
			else
				get[0]	= HPDI32_IO_MODE_DMA;
		}

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_mode
*
*	Purpose:
*
*		Implement the I/O Mode parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_mode(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_mode_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_overlap_enable_side
*
*	Purpose:
*
*		Implement the I/O Overlap Enable parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_overlap_enable_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_OVERLAP_ENABLE_DEFAULT;
		}
		else if (	(set == HPDI32_IO_OVERLAP_ENABLE_NO) ||
					(set == HPDI32_IO_OVERLAP_ENABLE_YES))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_IO_OVERLAP_ENABLE_YES;
		}
		else
		{
			set	= HPDI32_IO_OVERLAP_ENABLE_NO;
		}

		if (set != GSC_NO_CHANGE)
		{
			status	= gsc_io_overlap_enable(&io->gio.overlap, set, NULL);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		if (get)
		{
			status	= gsc_io_overlap_enable(&io->gio.overlap, GSC_NO_CHANGE, get);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_overlap_enable
*
*	Purpose:
*
*		Implement the I/O Overlap Enable parameter service. Validate and
*		apply the setting(s), as appropriate, then retrieve the current
*		setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_overlap_enable(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_overlap_enable_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_status_side
*
*	Purpose:
*
*		Implement the I/O Status parameter service for a single Tx/Rx side.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_status_side(dev_data_t* dev, dev_io_t* io, UL set, UL* get)
{
	U32	status	= (U32) GSC_SUCCESS;

	for (;;)	// A convenience loop.
	{
		// We don't lock access as we want status now,
		// not when any active I/O is done.

		// SET not supported for this parameter.

		if (get)
		{
			status	= gsc_io_status(&io->gio, get);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	io_status
*
*	Purpose:
*
*		Implement the I/O Status parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_status(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_status_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_pio_threshold_side
*
*	Purpose:
*
*		Implement the I/O PIO Threshold parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_pio_threshold_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_PIO_THRESHOLD_DEFAULT;
		}

		if (set != GSC_NO_CHANGE)
			io->gio.pio_threshold	= set;

		if (get)
			get[0]	= io->gio.pio_threshold;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_pio_threshold
*
*	Purpose:
*
*		Implement the I/O PIO Threshold parameter service. Validate and apply
*		the setting(s), as appropriate, then retrieve the current setting(s)
*		for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_pio_threshold(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_pio_threshold_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_timeout_side
*
*	Purpose:
*
*		Implement the I/O Timeout parameter service for a single Tx/Rx side.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_timeout_side(dev_data_t* dev, dev_io_t* io, UL set, UL* get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_TIMEOUT_DEFAULT;
		}
		else if (set > HPDI32_IO_TIMEOUT_MAX)
		{
			if (dev->strict_args)
			{
				status	= (U32) GSC_INVALID_DATA;
				break;
			}
			else
			{
				set	= HPDI32_IO_TIMEOUT_MAX;
			}
		}

		if (set != GSC_NO_CHANGE)
			io->gio.timeout_s	= set;

		if (get)
			get[0]	= io->gio.timeout_s;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_timeout
*
*	Purpose:
*
*		Implement the I/O Timeout parameter service. Validate and apply the
*		setting(s), as appropriate, then retrieve the current setting(s) for
*		the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_timeout(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_timeout_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	_io_single_cycle_side
*
*	Purpose:
*
*		Implement the I/O Single Cycle parameter service for a single Tx/Rx
*		side. Validate and apply the setting, as appropriate, then retrieve
*		the current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		io			The I/O structure for the Tx/Rx side to access.
*
*		set			The value to apply.
*
*		get			Retriev the current setting and record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_single_cycle_side(
	dev_data_t*	dev,
	dev_io_t*	io,
	UL			set,
	UL*			get)
{
	U32	status;
	U32	unlock;

	for (;;)	// A convenience loop.
	{
		status	= gsc_sem_lock(&io->gio.sem);
		unlock	= status == (U32) GSC_SUCCESS;

		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_IO_SINGLE_CYCLE_DEFAULT;
		}
		else if (	(set == HPDI32_IO_SINGLE_CYCLE_ABSENT) ||
					(set == HPDI32_IO_SINGLE_CYCLE_PRESENT))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}

		if (set != GSC_NO_CHANGE)
			io->single_cycle	= set;

		if (get)
			get[0]	= io->single_cycle;

		break;
	}

	if (unlock)
		gsc_sem_unlock(&io->gio.sem);

	return(status);
}



/******************************************************************************
*
*	Function:	io_single_cycle
*
*	Purpose:
*
*		Implement the I/O Single Cycle parameter service. Validate and apply
*		the setting(s), as appropriate, then retrieve the current setting(s)
*		for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_single_cycle(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	status	= _io_common_parameter_which(	dev,
											_io_single_cycle_side,
											which,
											set,
											get);
	return(status);
}



/******************************************************************************
*
*	Function:	hpdi32_io_wait
*
*	Purpose:
*
*		Wait for any one of the referenced I/O transfers to complete.
*
*	Arguments:
*
*		handle		The device to access.
*
*		which		The I/O transfer to wait on.
*
*		timeout_ms	Return if a completion doesn't occur within this number
*					of milliseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 GSC_EXPORT hpdi32_io_wait(void* handle, U32 which, U32 timeout_ms)
{
	dev_data_t*	dev;
	U32			status;

	status	= dev_access_lock(handle, &dev);

	if (status == (U32) GSC_SUCCESS)
	{
		status	= config_which_validate(dev, which, HPDI32_WHICH_TX_RX);

		if (status != (U32) GSC_SUCCESS)
		{
			dev_access_unlock(dev);
		}
		else
		{
			status	= gsc_io_wait(
							&dev->sem,
							(which & HPDI32_WHICH_TX) ? &dev->tx.gio : NULL,
							(which & HPDI32_WHICH_RX) ? &dev->rx.gio : NULL,
							timeout_ms);
		}
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_io_read
*
*	Purpose:
*
*		Perform a data read operation. This is the common routine called for
*		both blocking and overlapped I/O.
*
*	Arguments:
*
*		dev			The device to access.
*
*		buffer		The data read is put here.
*
*		bytes		The number of data bytes to read.
*
*		count		The actual number of bytes read is put here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_read(dev_data_t* dev, void* buffer, U32 bytes, U32* count)
{
	U32	samples;
	U32	status;

	// The I/O semaphore is already locked.
	dev->rx.gio.abort		= 0;
	dev->rx.gio.aborted		= 0;
	dev->rx.gio.buffer		= buffer;
	dev->rx.gio.xfer_buf	= buffer;
	dev->rx.gio.xfer_tot	= bytes;
	dev->rx.gio.xfered_tot	= 0;

	samples	= bytes / (dev->rx.gio.data_size / 8);

	if (samples <= dev->rx.gio.pio_threshold)
	{
		// Revert to PIO.
		dev->rx.gio.dma_need	= 0;
		dev->rx.gio.func		= dev->rx.cache.pio_funcs;
	}
	else
	{
		// Use the selected mode.
		dev->rx.gio.dma_need	= dev->rx.cache.dma_need;
		dev->rx.gio.func		= dev->rx.cache.mode_funcs;
	}

	status	= gsc_io_transfer_work(&dev->rx.gio);

	if (count)
		count[0]	= dev->rx.gio.xfered_tot;

	return(status);
}



/******************************************************************************
*
*	Function:	hpdi32_read
*
*	Purpose:
*
*		Perform a data read operation.
*
*	Arguments:
*
*		handle		The device to access.
*
*		buffer		The data read is put here.
*
*		bytes		The number of data bytes to read.
*
*		count		The actual number of bytes read is put here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 GSC_EXPORT hpdi32_read(void* handle, void* buffer, U32 bytes, U32* count)
{
	dev_data_t*	dev;
	U32			status;

	status	= dev_access_lock(handle, &dev);

	if (status == (U32) GSC_SUCCESS)
	{
		status	= gsc_io_start_transfer(&dev->sem,
										&dev->rx.gio,
										_io_read,
										buffer,
										bytes,
										count);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_io_write
*
*	Purpose:
*
*		Perform a data write operation. This is the common routine called for
*		both blocking and overlapped I/O.
*
*	Arguments:
*
*		dev			The device to access.
*
*		buffer		The data is written from here.
*
*		bytes		The number of data bytes to write.
*
*		count		The actual number of bytes written is put here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

static U32 _io_write(dev_data_t* dev, void* buffer, U32 bytes, U32* count)
{
	U32	samples;
	U32	status;

	// The I/O semaphore is already locked.
	dev->tx.gio.abort		= 0;
	dev->tx.gio.aborted		= 0;
	dev->tx.gio.buffer		= buffer;
	dev->tx.gio.xfer_buf	= buffer;
	dev->tx.gio.xfer_tot	= bytes;
	dev->tx.gio.xfered_tot	= 0;

	samples	= bytes / (dev->tx.gio.data_size / 8);

	if (samples <= dev->tx.gio.pio_threshold)
	{
		// Revert to PIO.
		dev->tx.gio.dma_need	= 0;
		dev->tx.gio.func		= dev->tx.cache.pio_funcs;
	}
	else
	{
		// Use the selected mode.
		dev->tx.gio.dma_need	= dev->tx.cache.dma_need;
		dev->tx.gio.func		= dev->tx.cache.mode_funcs;
	}

	status	= gsc_io_transfer_work(&dev->tx.gio);

	if (count)
		count[0]	= dev->tx.gio.xfered_tot;

	return(status);
}



/******************************************************************************
*
*	Function:	hpdi32_write
*
*	Purpose:
*
*		Perform a data write operation.
*
*	Arguments:
*
*		handle		The device to access.
*
*		buffer		The source for the data to write.
*
*		bytes		The number of data bytes to write.
*
*		count		The actual number of bytes written is put here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 GSC_EXPORT hpdi32_write(void* handle, const void* buffer, U32 bytes, U32* count)
{
	dev_data_t*	dev;
	U32			status;

	status	= dev_access_lock(handle, &dev);

	if (status == (U32) GSC_SUCCESS)
	{
		status	= gsc_io_start_transfer(&dev->sem,
										&dev->tx.gio,
										_io_write,
										(void*) buffer,
										bytes,
										count);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	io_auto_start
*
*	Purpose:
*
*		Initiate automatic data transfer.
*
*	Arguments:
*
*		io			The common I/O parameter structure.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 io_auto_start(gsc_io_t* io)
{
	U32	status;

	// This is for Tx only.
	status	= GSC_REG_MOD(	io->dev,
							HPDI32_BCR,
							HPDI32_BCR_TX_START,
							HPDI32_BCR_TX_START);
	return(status);
}



/******************************************************************************
*
*	Function:	io_init
*
*	Purpose:
*
*		Initialize the given I/O structure to the devices initial open()
*		state. The semaphore is assumed to already be in use when we are
*		called, so we'll work without trying to gain exclusive access.
*
*	Arguments:
*
*		dev			The device being accessed.
*
*		dio			The device I/O structure to initialize.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 io_init(dev_data_t* dev, dev_io_t* dio)
{
	U32	io_stat;
	U32	stat;
	U32	status	= (U32) GSC_SUCCESS;

#if (HPDI32_IO_MODE_DEFAULT != HPDI32_IO_MODE_DMDMA)
	#error "Default I/O mode not supported here."
#endif

	stat	= gsc_io_init(&dio->gio);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_mem_alloc(&dio->mem, HPDI32_IO_BUFFER_SIZE_DEFAULT, NULL);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	dio->use_pio		= 0;
	dio->single_cycle	= HPDI32_IO_SINGLE_CYCLE_DEFAULT;

	if (dio->gio.is_tx)
	{
		dio->gio.auto_start		= HPDI32_TX_AUTO_START_DEFAULT;
		dio->gio.dma_priority	= HPDI32_IO_DMA_PRIORITY_TX_DEFAULT;
	}
	else
	{
		dio->gio.auto_start		= 0;
		dio->gio.dma_priority	= HPDI32_IO_DMA_PRIORITY_RX_DEFAULT;
	}

	dio->gio.data_lane			= 0;
	dio->gio.data_size			= HPDI32_IO_DATA_SIZE_DEFAULT;
	dio->gio.dma_control_mode	= HPDI32_IO_DMA_CONTROL_MODE_DEFAULT;
	dio->gio.dma_demand			= 1;	// DMDMA
	dio->gio.dma_dynamic		= HPDI32_IO_DMA_CHANNEL_SEL_TX_DEFAULT;
	dio->gio.dma_enable			= 1;	// DMA
	dio->gio.pio_threshold		= HPDI32_IO_PIO_THRESHOLD_DEFAULT;
	dio->gio.reg				= HPDI32_FDR;
	dio->gio.timeout_s			= HPDI32_IO_TIMEOUT_DEFAULT;

	stat	= gsc_dma_channel_close(&dev->dma, &dio->gio.dma_channel);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_io_overlap_enable(&dio->gio.overlap,
									HPDI32_IO_OVERLAP_ENABLE_DEFAULT,
									NULL);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	if (dio->gio.is_tx)
		io_tx_funcs_select(dev);
	else
		_io_rx_funcs_select(dev);

	// Initialize the I/O notification stuff.

	if (dio->gio.is_tx)
		io_stat	= GSC_IO_STATUS_TX;
	else
		io_stat	= 0;

	stat	= gsc_io_callback_config(
				&dio->gio,
				0,				// id: index
				NULL,			// function
				dev,			// arg1: device handle
				io_stat,		// arg2: I/O status
				0,				// arg3: user argument
				NULL);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	return(status);
}



/******************************************************************************
*
*	Function:	io_tx_funcs_select
*
*	Purpose:
*
*		Perform a check to see which I/O functions will be needed when a write
*		is requested.
*
*	Arguments:
*
*		dev			The device to access.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void io_tx_funcs_select(dev_data_t* dev)
{
	if (dev->tx.gio.auto_start)
		dev->tx.cache.pio_funcs	= &io_pio_write_ase_funcs;
	else
		dev->tx.cache.pio_funcs	= &io_pio_write_asd_funcs;

	if (dev->tx.gio.dma_enable == 0)
	{
		// PIO
		dev->tx.cache.dma_need	= 0;

		if (dev->tx.gio.auto_start)
			dev->tx.cache.mode_funcs	= &io_pio_write_ase_funcs;
		else
			dev->tx.cache.mode_funcs	= &io_pio_write_asd_funcs;
	}
	else if (dev->tx.gio.dma_demand)
	{
		// Demand Mode DMA
		dev->tx.cache.dma_need	= GSC_DMA_FLAG_DMDMA_TX;

		if (dev->cache.single_cycle_disable)
		{
			// The board performs Single Cycles when the Tx FIFO rises to Almost Full
			// and firmware supports Single Cycle Disable.

			if (dev->tx.gio.auto_start)
				dev->tx.cache.mode_funcs	= &io_dmdma_write_scd_ase_funcs;
			else
				dev->tx.cache.mode_funcs	= &io_dmdma_write_scd_asd_funcs;
		}
		else if (dev->tx.single_cycle == HPDI32_IO_SINGLE_CYCLE_PRESENT)
		{
			// The board performs Single Cycles when the Tx FIFO rises to Almost Full
			// but firmware doesn't support Single Cycle Disable.

			if (dev->tx.gio.auto_start)
				dev->tx.cache.mode_funcs	= &io_dmdma_write_scp_ase_funcs;
			else
				dev->tx.cache.mode_funcs	= &io_dmdma_write_scp_asd_funcs;
		}
		else
		{
			// The board doesn't perform Single Cycles when the Tx FIFO rises
			// to Almost Full.

			if (dev->tx.gio.auto_start)
				dev->tx.cache.mode_funcs	= &io_dmdma_write_sca_ase_funcs;
			else
				dev->tx.cache.mode_funcs	= &io_dmdma_write_sca_asd_funcs;
		}
	}
	else
	{
		// Non Demand Mode DMA
		dev->tx.cache.dma_need	= GSC_DMA_FLAG_BLDMA_TX;

		if (dev->tx.gio.dma_control_mode == HPDI32_IO_DMA_CONTROL_MODE_AUTOMATIC)
		{
			// Safe Blind DMA (API breaks request into smaller chinks).

			if (dev->tx.gio.auto_start)
				dev->tx.cache.mode_funcs	= &io_dma_write_auto_ase_funcs;
			else
				dev->tx.cache.mode_funcs	= &io_dma_write_auto_asd_funcs;
		}
		else	// HPDI32_IO_DMA_CONTROL_MODE_MANUAL
		{
			// Risky Blind DMA (application must control size of requests).

			if (dev->tx.gio.auto_start)
				dev->tx.cache.mode_funcs	= &io_dma_write_manual_ase_funcs;
			else
				dev->tx.cache.mode_funcs	= &io_dma_write_manual_asd_funcs;
		}
	}
}


