// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/dev_misc.c $
// $Rev: 30442 $
// $Date: 2014-10-07 16:31:00 -0500 (Tue, 07 Oct 2014) $

#include "dev_misc.h"

#include "gsc_reg.h"

#include "dev_config.h"
#include "dev_reg.h"



/******************************************************************************
*
*	Function:	misc_board_jumpers
*
*	Purpose:
*
*		Implement the miscelaneous Board Jumpers parameter service. Validate
*		and apply the setting, as appropriate, then retrieve the current
*		setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_board_jumpers(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	bsr;
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET is not supported by this parameter.

		status	= GSC_REG_READ(dev, HPDI32_BSR, &bsr);
		if (status != (U32) GSC_SUCCESS)	break;

		if (dev->cache.user_jumpers)
			set	= (bsr >> 16) & 0x3;
		else
			set	= 0;

		if (get)
			get[0]	= set;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_favor_tx
*
*	Purpose:
*
*		Implement the miscallaneous Favor Tx parameter service. Validate and
*		apply the setting, as appropriate, then retrieve the current
*		setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_favor_tx(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_MISC_FAVOR_TX_DEFAULT;
		}
		else if (	(set == HPDI32_MISC_FAVOR_TX_ENABLE) ||
					(set == HPDI32_MISC_FAVOR_TX_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_MISC_FAVOR_TX_ENABLE;
		}
		else
		{
			set	= HPDI32_MISC_FAVOR_TX_DISABLE;
		}

		if (set != GSC_NO_CHANGE)
			dev->favor_tx	= set;

		if (get)
			get[0]	= dev->favor_tx;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_features
*
*	Purpose:
*
*		Implement the miscallaneous Features parameter service. Validate and
*		apply the setting, as appropriate, then retrieve the current
*		setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			This gives the feature to check for.
*
*		get			Support for the given feature is recorded here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_features(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status	= (U32) GSC_SUCCESS;
	U32	value	= HPDI32_MISC_FEATURES_ABSENT;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET - set argument specifies the feature to GET

		// GET

		switch (set)
		{
			default:	if (dev->strict_args)
							status	= (U32) GSC_INVALID_DATA;

						break;

			case GSC_DEFAULT:
			case GSC_NO_CHANGE:

						break;

			case HPDI32_MISC_FEATURES_1_CYCLE_DISABLE:

						if (dev->cache.single_cycle_disable)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_COUNT:

						value	= HPDI32_MISC_FEATURES_LAST_INDEX + 1;
						break;

			case HPDI32_MISC_FEATURES_FSR:

						if (dev->cache.frr & HPDI32_FRR_FSR)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_FIFO_SIZE:

						if (dev->cache.fsr & HPDI32_FSR_TFSR_RFSR)
							value	= (U32) HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_ICR:

						if (dev->cache.fsr & HPDI32_FSR_IELR_IHLR)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_GPIO_0_5:

						if (dev->cache.fsr & HPDI32_FSR_GPIO_0_5)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_GPIO_6:
			case HPDI32_MISC_FEATURES_TX_AUTO_STOP:

						if (dev->cache.fsr & HPDI32_FSR_GPIO6_TXAS)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_DMA_CH1:

						if (dev->cache.fsr & HPDI32_FSR_DMDMA_CH1)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_OVR_UNDR_RUN:

						if (dev->cache.fsr & HPDI32_FSR_OVR_UNDR_RUN)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_USER_JUMPERS:

						if (dev->cache.user_jumpers)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;

			case HPDI32_MISC_FEATURES_GPIO_D32:

						if (dev->cache.fsr & HPDI32_FSR_GPIO_D32)
							value	= HPDI32_MISC_FEATURES_PRESENT;

						break;
		}

		if (status != (U32) GSC_SUCCESS)
			break;

		if (get)
			get[0]	= value;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_map_gsc_regs
*
*	Purpose:
*
*		Implement the miscallaneous service for mapping the GSC register
*		block to application space. Validate and apply the setting, as
*		appropriate, then retrieve the current setting for the caller, also
*		as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_map_gsc_regs(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	rx_lock;
	U32	status;
	U32	tx_lock;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_MISC_MAP_GSC_REGS_DEFAULT;
		}
		else if (	(set == HPDI32_MISC_MAP_GSC_REGS_ENABLE) ||
					(set == HPDI32_MISC_MAP_GSC_REGS_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_MISC_MAP_GSC_REGS_ENABLE;
		}
		else
		{
			set	= HPDI32_MISC_MAP_GSC_REGS_DISABLE;
		}

		if (set != GSC_NO_CHANGE)
		{
			if (set == HPDI32_MISC_MAP_GSC_REGS_ENABLE)
			{
				status	= gsc_bar_map(&dev->gsc_regs);
			}
			else
			{
				rx_lock	= gsc_sem_lock(&dev->rx.gio.sem);
				tx_lock	= gsc_sem_lock(&dev->tx.gio.sem);

				if ((rx_lock == (U32) GSC_SUCCESS) &&
					(tx_lock == (U32) GSC_SUCCESS))
				{
					status	= gsc_bar_unmap(&dev->gsc_regs);
				}

				if (rx_lock == (U32) GSC_SUCCESS)
					gsc_sem_unlock(&dev->rx.gio.sem);

				if (tx_lock == (U32) GSC_SUCCESS)
					gsc_sem_unlock(&dev->tx.gio.sem);

			}

			if (status != (U32) GSC_SUCCESS)
				break;
		}

		if (get)
		{
			if (dev->gsc_regs.ptr)
				get[0]	= HPDI32_MISC_MAP_GSC_REGS_ENABLE;
			else
				get[0]	= HPDI32_MISC_MAP_GSC_REGS_DISABLE;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_map_gsc_regs_ptr
*
*	Purpose:
*
*		Implement the miscallaneous service for retrieving the pointer to the
*		memory mapped GSC register block. Validate and apply the setting, as
*		appropriate, then retrieve the current setting for the caller, also
*		as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_map_gsc_regs_ptr(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET is not supported for this parameter.

		if (get)
			get[0]	= (UL) dev->gsc_regs.ptr;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_map_plx_regs
*
*	Purpose:
*
*		Implement the miscallaneous service for mapping the PLX feature set
*		register block to application space. Validate and apply the setting,
*		as appropriate, then retrieve the current setting for the caller,
*		also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_map_plx_regs(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_MISC_MAP_PLX_REGS_DEFAULT;
		}
		else if (	(set == HPDI32_MISC_MAP_PLX_REGS_ENABLE) ||
					(set == HPDI32_MISC_MAP_PLX_REGS_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_MISC_MAP_PLX_REGS_ENABLE;
		}
		else
		{
			set	= HPDI32_MISC_MAP_PLX_REGS_DISABLE;
		}

		if (set != GSC_NO_CHANGE)
		{
			if (set == HPDI32_MISC_MAP_PLX_REGS_ENABLE)
				status	= gsc_bar_map(&dev->plx_regs);
			else
				status	= gsc_bar_unmap(&dev->plx_regs);

			if (status != (U32) GSC_SUCCESS)
				break;
		}

		if (get)
		{
			if (dev->plx_regs.ptr)
				get[0]	= HPDI32_MISC_MAP_PLX_REGS_ENABLE;
			else
				get[0]	= HPDI32_MISC_MAP_PLX_REGS_DISABLE;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_pci_bus_width
*
*	Purpose:
*
*		Implement the miscallaneous service for retrieving the us width of
*		the current device. Validate and apply the setting, as appropriate,
*		then retrieve the current setting for the caller, also as
*		appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_pci_bus_width(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET is not supported for this parameter.

		if (get)
		{
			if (dev->cache.bus_width == 64)
				get[0]	= HPDI32_MISC_PCI_BUS_WIDTH_64;
			else
				get[0]	= HPDI32_MISC_PCI_BUS_WIDTH_32;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_strict_arguments
*
*	Purpose:
*
*		Implement the miscallaneous Strict Argument parameter service.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_strict_arguments(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_MISC_STRICT_ARGUMENTS_DEFAULT;
		}
		else if (	(set == HPDI32_MISC_STRICT_ARGUMENTS_ENABLE) ||
					(set == HPDI32_MISC_STRICT_ARGUMENTS_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_MISC_STRICT_ARGUMENTS_ENABLE;
		}
		else
		{
			set	= HPDI32_MISC_STRICT_ARGUMENTS_DISABLE;
		}

		if (set != GSC_NO_CHANGE)
			dev->strict_args	= set;

		if (get)
			get[0]	= dev->strict_args;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_strict_config
*
*	Purpose:
*
*		Implement the miscallaneous Strict Configuration parameter service.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting(s) for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_strict_config(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_MISC_STRICT_CONFIG_DEFAULT;
		}
		else if (	(set == HPDI32_MISC_STRICT_CONFIG_ENABLE) ||
					(set == HPDI32_MISC_STRICT_CONFIG_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_MISC_STRICT_CONFIG_ENABLE;
		}
		else
		{
			set	= HPDI32_MISC_STRICT_CONFIG_DISABLE;
		}

		if (set != GSC_NO_CHANGE)
			dev->strict_config	= set;

		if (get)
			get[0]	= dev->strict_config;

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	misc_tx_rx_tri_state
*
*	Purpose:
*
*		Implement the miscelaneous Tx/Rx Tri-State parameter service.
*		Validate and apply the setting, as appropriate, then retrieve the
*		current setting for the caller, also as appropriate.
*
*	Arguments:
*
*		dev			The device to access.
*
*		which		A field of bits specifying the subjects to act on.
*
*		set			The value to apply.
*
*		get			Retriev the current setting an record it here, if not
*					NULL.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		A status reflecting the error encountered.
*
******************************************************************************/

U32 misc_tx_rx_tri_state(dev_data_t* dev, U32 which, UL set, UL* get)
{
	U32	bcr;
	U32	status;

	for (;;)	// A convenience loop.
	{
		status	= config_which_validate(dev, which, 0);
		if (status != (U32) GSC_SUCCESS)	break;

		// SET

		if (set == GSC_NO_CHANGE)
		{
		}
		else if (set == GSC_DEFAULT)
		{
			set	= HPDI32_MISC_TX_RX_TRI_STATE_DEFAULT;
		}
		else if (	(set == HPDI32_MISC_TX_RX_TRI_STATE_ENABLE) ||
					(set == HPDI32_MISC_TX_RX_TRI_STATE_DISABLE))
		{
		}
		else if (dev->strict_args)
		{
			status	= (U32) GSC_INVALID_DATA;
			break;
		}
		else if (set)
		{
			set	= HPDI32_MISC_TX_RX_TRI_STATE_ENABLE;
		}
		else
		{
			set	= HPDI32_MISC_TX_RX_TRI_STATE_DISABLE;
		}

		if (set != GSC_NO_CHANGE)
		{
			if (set == HPDI32_MISC_TX_RX_TRI_STATE_ENABLE)
				set	= HPDI32_BCR_TEST_MODE;
			else
				set	= 0;

			status	= GSC_REG_MOD(dev, HPDI32_BCR, set, HPDI32_BCR_TEST_MODE);
			if (status != (U32) GSC_SUCCESS)	break;
		}

		if (get)
		{
			status	= GSC_REG_READ(dev, HPDI32_BCR, &bcr);
			if (status != (U32) GSC_SUCCESS)	break;

			get[0]	= (bcr & HPDI32_BCR_TEST_MODE)
					? HPDI32_MISC_TX_RX_TRI_STATE_ENABLE
					: HPDI32_MISC_TX_RX_TRI_STATE_DISABLE;
		}

		break;
	}

	return(status);
}


