// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/gsc_reg.c $
// $Rev: 30377 $
// $Date: 2014-09-18 12:09:04 -0500 (Thu, 18 Sep 2014) $

#include "gsc_reg.h"

#include "gsc_drv.h"



/******************************************************************************
*
*	Function:	gsc_reg_io_mod
*
*	Purpose:
*
*		Perform a read-modify-write on an I/O mapped register.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		index		The index of the BAR to access.
*
*		r_size		The size of region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to write
*
*		mask		The set of bits to modify. Excess bits are ignored.
*					If a bit is set here, then the register bits get
*					changed.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_io_mod(	gsc_device_t*	device,
					U8				index,
					U32				r_size,
					U32				reg,
					U32				value,
					U32				mask)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				(size == 3) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_io_mod(	device,
										(S8) index,
										offset,
										size,
										value,
										mask);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_io_read
*
*	Purpose:
*
*		Read an I/O mapped register.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		index		The index of the BAR to access.
*
*		r_size		The size of region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to read is put here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_io_read(gsc_device_t* device, U8 index, U32 r_size, U32 reg, U32* value)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_io_read(device, (S8) index, offset, size, value);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_io_write
*
*	Purpose:
*
*		Write to an I/O mapped register.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		index		The index of the BAR to access.
*
*		r_size		The size of region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to write.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_io_write(gsc_device_t* device, U8 index, U32 r_size, U32 reg, U32 value)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_io_write(device, (S8) index, offset, size, value);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_mapped_mod
*
*	Purpose:
*
*		Perform a read-modify-write on a directly accessible memory based
*		register.
*
*	Arguments:
*
*		r_base		The base address of the region where the register resides.
*
*		r_size		The size of the region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to write
*
*		mask		The set of bits to modify. Excess bits are ignored.
*					If a bit is set here, then the register bits get
*					changed.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_mapped_mod(UL r_base, U32 r_size, U32 reg, U32 value, U32 mask)
{
	U16		offset;
	U16*	pu16;
	U32*	pu32;
	U8*		pu8;
	U8		size;
	U32		status;
	U16		u16;
	U32		u32;
	U8		u8;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if ((size == 4) && ((offset % 4) == 0))
	{
		pu32	= (U32*) (r_base + offset);
		u32		= pu32[0];
		u32		= (u32 & ~mask) | (value & mask);
		pu32[0]	= u32;
		status	= (U32) GSC_SUCCESS;
	}
	else if ((size == 2) && ((offset % 2) == 0))
	{
		pu16	= (U16*) (r_base + offset);
		u16		= pu16[0];
		u16		= (U16) ((u16 & ~mask) | (value & mask));
		pu16[0]	= u16;
		status	= (U32) GSC_SUCCESS;
	}
	else if (size == 1)
	{
		pu8		= (U8*) (r_base + offset);
		u8		= pu8[0];
		u8		= (U8) ((u8 & ~mask) | (value & mask));
		pu8[0]	= u8;
		status	= (U32) GSC_SUCCESS;
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_mapped_read
*
*	Purpose:
*
*		Read from a directly accessible memory based register.
*
*	Arguments:
*
*		r_base		The base address of the region where the register resides.
*
*		r_size		The size of the region where the register resides.
*
*		reg			The register to access.
*
*		value		The value read is put here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_mapped_read(UL r_base, U32 r_size, U32 reg, U32* value)
{
	U16		offset;
	U16*	pu16;
	U32*	pu32;
	U8*		pu8;
	U8		size;
	U32		status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if ((size == 4) && ((offset % 4) == 0))
	{
		pu32		= (U32*) (r_base + offset);
		value[0]	= pu32[0];
		status		= (U32) GSC_SUCCESS;
	}
	else if ((size == 2) && ((offset % 2) == 0))
	{
		pu16		= (U16*) (r_base + offset);
		value[0]	= pu16[0];
		status		= (U32) GSC_SUCCESS;
	}
	else if (size == 1)
	{
		pu8			= (U8*) (r_base + offset);
		value[0]	= pu8[0];
		status		= (U32) GSC_SUCCESS;
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_mapped_write
*
*	Purpose:
*
*		Write to a directly accessible memory based register.
*
*	Arguments:
*
*		r_base		The base address of the region where the register resides.
*
*		r_size		The size of the region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to write.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_mapped_write(UL r_base, U32 r_size, U32 reg, U32 value)
{
	U16				offset;
	volatile U16*	pu16;
	volatile U32*	pu32;
	volatile U8*	pu8;
	U8				size;
	U32				status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if ((size == 4) && ((offset % 4) == 0))
	{
		pu32	= (U32*) (r_base + offset);
		pu32[0]	= value;
		status	= (U32) GSC_SUCCESS;
	}
	else if ((size == 2) && ((offset % 2) == 0))
	{
		pu16	= (U16*) (r_base + offset);
		pu16[0]	= (U16) value;
		status	= (U32) GSC_SUCCESS;
	}
	else if (size == 1)
	{
		pu8		= (U8*) (r_base + offset);
		pu8[0]	= (U8) value;
		status	= (U32) GSC_SUCCESS;
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_mem_mod
*
*	Purpose:
*
*		Perform a read-modify-write on a NON-directly accessible memory
*		mapped register.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		index		The index of the BAR to access.
*
*		r_size		The size of the region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to write
*
*		mask		The set of bits to modify. Excess bits are ignored.
*					If a bit is set here, then the register bits get
*					changed.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_mem_mod(gsc_device_t*	device,
					U8				index,
					U32				r_size,
					U32				reg,
					U32				value,
					U32				mask)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_mem_mod(	device,
										(S8) index,
										offset,
										size,
										value,
										mask);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_mem_read
*
*	Purpose:
*
*		Read from a NON-directly accessible memory mapped register.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		index		The index of the BAR to access.
*
*		r_size		The size of the region where the register resides.
*
*		reg			The register to access.
*
*		value		The value read goes here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_mem_read(gsc_device_t* device, U8 index, U32 r_size, U32 reg, U32* value)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_mem_read(device, (S8) index, offset, size, value);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_mem_write
*
*	Purpose:
*
*		Write to a NON-directly accessible memory mapped register.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		index		The index of the BAR to access.
*
*		r_size		The size of the region where the register resides.
*
*		reg			The register to access.
*
*		value		The value to write.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_mem_write(gsc_device_t* device, U8 index, U32 r_size, U32 reg, U32 value)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= (U16) r_size)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_mem_write(device, (S8) index, offset, size, value);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	gsc_reg_pci_read
*
*	Purpose:
*
*		Read a PCI register from the referenced device.
*
*	Arguments:
*
*		device		The data used to access the device.
*
*		reg			The register to access.
*
*		value		The value read is recorded here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
******************************************************************************/

U32 gsc_reg_pci_read(gsc_device_t* device, U32 reg, U32* value)
{
	U16	offset;
	U8	size;
	U32	status;

	offset	= (U16) GSC_REG_OFFSET(reg);
	size	= (U8) GSC_REG_SIZE(reg);

	if ((offset + size - 1) >= 256)
	{
		status	= (U32) GSC_INVALID_DATA;
	}
	else if (	((size == 4) && ((offset % 4) == 0)) ||
				(size == 3) ||
				((size == 2) && ((offset % 2) == 0)) ||
				(size == 1))
	{
		status	= gsc_drv_reg_pci_read(device, offset, size, value);
	}
	else
	{
		status	= (U32) GSC_INVALID_DATA;
	}

	return(status);
}



