// This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL).
// Any Device - Driver - Linux
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/driver/reg.c $
// $Rev: 23671 $
// $Date: 2013-10-31 16:23:27 -0500 (Thu, 31 Oct 2013) $

#include "main.h"



/******************************************************************************
*
*	Function:	_reg_io_mod
*
*	Purpose:
*
*		Perform a read-modify-write on an I/O mapped register. Sizes may be
*		1, 2 or 4 bytes. Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_io_mod(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	dev_bar_t*		bar;
	u8				byte;
	u32				dword;
	unsigned int	port;
	int				status;
	u16				word;

	reg->status	= GSC_SUCCESS;
	bar			= &dev->bars.bar[reg->bar];
	port		= ((unsigned int) (unsigned long) bar->vaddr) + reg->offset;

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= bar->size)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			byte	= inb(port);
			byte	= (byte & ~reg->mask) | (reg->value & reg->mask);
			outb(byte, port);
			break;

		case 2:

			if (((reg->offset + 2) >= bar->size) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			word	= inw(port);
			word	= (word & ~reg->mask) | (reg->value & reg->mask);
			outw(word, port);
			break;

		case 4:

			if (((reg->offset + 4) >= bar->size) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			dword	= inl(port);
			dword	= (dword & ~reg->mask) | (reg->value & reg->mask);
			outl(dword, port);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_io_read
*
*	Purpose:
*
*		Read from an I/O mapped register. Sizes may be 1, 2 or 4 bytes.
*		Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_io_read(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	dev_bar_t*		bar;
	unsigned int	port;
	int				status;

	reg->status	= GSC_SUCCESS;
	bar			= &dev->bars.bar[reg->bar];
	port		= ((unsigned int) (unsigned long) bar->vaddr) + reg->offset;

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= bar->size)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			reg->value	= (u8) inb(port);
			break;

		case 2:

			if (((reg->offset + 2) >= bar->size) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			reg->value	= (u16) inw(port);
			break;

		case 4:

			if (((reg->offset + 4) >= bar->size) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			reg->value	= (u32) inl(port);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_io_write
*
*	Purpose:
*
*		Write to an I/O mapped register. Sizes may be 1, 2 or 4 bytes.
*		Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_io_write(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	dev_bar_t*		bar;
	unsigned int	port;
	int				status;

	reg->status	= GSC_SUCCESS;
	bar			= &dev->bars.bar[reg->bar];
	port		= ((unsigned int) (unsigned long) bar->vaddr) + reg->offset;

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= bar->size)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			outb((u8) reg->value, port);
			break;

		case 2:

			if (((reg->offset + 2) >= bar->size) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			outw((u16) reg->value, port);
			break;

		case 4:

			if (((reg->offset + 4) >= bar->size) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			outl((u32) reg->value, port);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_mem_mod
*
*	Purpose:
*
*		Perform a read-modify-write on a memory mapped register. Sizes may be
*		1, 2 or 4 bytes. Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_mem_mod(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	dev_bar_t*	bar;
	u8			byte;
	u32			dword;
	int			status;
	VADDR_T		vaddr;
	u16			word;

	reg->status	= GSC_SUCCESS;
	bar			= &dev->bars.bar[reg->bar];
	vaddr		= (VADDR_T) (((unsigned long) bar->vaddr) + reg->offset);

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= bar->size)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			byte	= readb(vaddr);
			byte	= (byte & ~reg->mask) | (reg->value & reg->mask);
			writeb(byte, vaddr);
			break;

		case 2:

			if (((reg->offset + 2) >= bar->size) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			word	= readw(vaddr);
			word	= (word & ~reg->mask) | (reg->value & reg->mask);
			writew(word, vaddr);
			break;

		case 4:

			if (((reg->offset + 4) >= bar->size) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			dword	= readl(vaddr);
			dword	= (dword & ~reg->mask) | (reg->value & reg->mask);
			writel(dword, vaddr);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_mem_read
*
*	Purpose:
*
*		Read from a memory mapped register. Sizes may be 1, 2 or 4 bytes.
*		Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_mem_read(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	dev_bar_t*	bar;
	int			status;
	VADDR_T		vaddr;

	reg->status	= GSC_SUCCESS;
	bar			= &dev->bars.bar[reg->bar];
	vaddr		= (VADDR_T) (((unsigned long) bar->vaddr) + reg->offset);

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= bar->size)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			reg->value	= (u8) readb(vaddr);
			break;

		case 2:

			if (((reg->offset + 2) >= bar->size) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			reg->value	= (u16) readw(vaddr);
			break;

		case 4:

			if (((reg->offset + 4) >= bar->size) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			reg->value	= (u32) readl(vaddr);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_mem_write
*
*	Purpose:
*
*		Write to a memory mapped register. Sizes may be 1, 2 or 4 bytes.
*		Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_mem_write(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	dev_bar_t*	bar;
	int			status;
	VADDR_T		vaddr;

	reg->status	= GSC_SUCCESS;
	bar			= &dev->bars.bar[reg->bar];
	vaddr		= (VADDR_T) (((unsigned long) bar->vaddr) + reg->offset);

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= bar->size)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			writeb(reg->value, vaddr);
			break;

		case 2:

			if (((reg->offset + 2) >= bar->size) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			writew(reg->value, vaddr);
			break;

		case 4:

			if (((reg->offset + 4) >= bar->size) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			writel(reg->value, vaddr);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_mod
*
*	Purpose:
*
*		Perform a read-modify-write on a PCI Sizes may be 1, 2 or 4 bytes.
*		Alignment must be per the size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_mod(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	u8			byte;
	u32			dword;
	int			status;
	u16			word;

	reg->status	= GSC_SUCCESS;

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= 256)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			pci_read_config_byte(	dev->pci,
									(unsigned long) reg->offset,
									&byte);
			byte	= (byte & ~reg->mask) | (reg->value & reg->mask);
			pci_write_config_byte(	dev->pci,
									(unsigned long) reg->offset,
									byte);
			break;

		case 2:

			if (((reg->offset + 2) >= 256) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			pci_read_config_word(	dev->pci,
									(unsigned long) reg->offset,
									&word);
			word	= (word & ~reg->mask) | (reg->value & reg->mask);
			pci_write_config_word(	dev->pci,
									(unsigned long) reg->offset,
									word);
			break;

		case 4:

			if (((reg->offset + 4) >= 256) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			pci_read_config_dword(	dev->pci,
									(unsigned long) reg->offset,
									&dword);
			dword	= (dword & ~reg->mask) | (reg->value & reg->mask);
			pci_write_config_dword(	dev->pci,
									(unsigned long) reg->offset,
									dword);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_read_1
*
*	Purpose:
*
*		Read a PCI register that is one byte long.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_read_1(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;
	u8	v;

	if (reg->offset <= 255)
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_byte(	dev->pci,
								(unsigned long) reg->offset,
								&v);
		reg->value	= v;
	}
	else
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_OFFSET;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_read_2
*
*	Purpose:
*
*		Read a PCI register that is two bytes long. Any alignment is supported.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_read_2(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;
	u16	v1;
	u16	v2;

	if (((reg->offset & 1) == 0) && (reg->offset <= 254))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_word(	dev->pci,
								(unsigned long) reg->offset,
								&v1);
		reg->value	= v1;
	}
	else if (((reg->offset & 1) == 1) && (reg->offset <= 253))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_word(	dev->pci,
								(unsigned long) reg->offset - 1,
								&v1);
		pci_read_config_word(	dev->pci,
								(unsigned long) reg->offset + 1,
								&v2);
		reg->value	= ((v2 & 0x00FF) << 8) | ((v1 & 0xFF00) >> 8);
	}
	else
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_OFFSET;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_read_3
*
*	Purpose:
*
*		Read a PCI register that is three bytes long. Any alignment is
*		supported.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_read_3(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;
	u32	v1;
	u32	v2;

	if (((reg->offset & 3) == 0) && (reg->offset <= 252))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset,
								&v1);
		reg->value	= v1 & 0x00FFFFFF;
	}
	else if (((reg->offset & 3) == 1) && (reg->offset <= 252))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset - 1,
								&v1);
		reg->value	= (v1 & 0xFFFFFF00) >> 8;
	}
	else if (((reg->offset & 3) == 2) && (reg->offset <= 248))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset - 2,
								&v1);
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset + 2,
								&v2);
		reg->value	= ((v1 & 0xFFFF0000) >> 16) | ((v2 & 0x000000FF) << 16);
	}
	else if (((reg->offset & 3) == 3) && (reg->offset <= 248))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset - 3,
								&v1);
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset + 1,
								&v2);
		reg->value	= ((v1 & 0xFF000000) >> 24) | ((v2 & 0x0000FFFF) << 8);
	}
	else
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_OFFSET;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_read_4
*
*	Purpose:
*
*		Read a PCI register that is four bytes long. Any alignment is
*		supported.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_read_4(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;
	u32	v1;
	u32	v2;

	if (((reg->offset & 3) == 0) && (reg->offset <= 252))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset,
								&v1);
		reg->value	= v1;
	}
	else if (((reg->offset & 3) == 1) && (reg->offset <= 248))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset - 1,
								&v1);
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset + 3,
								&v2);
		reg->value	= ((v1 & 0xFFFFFF00) >> 8) | ((v2 & 0x000000FF) << 24);
	}
	else if (((reg->offset & 3) == 2) && (reg->offset <= 248))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset - 2,
								&v1);
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset + 2,
								&v2);
		reg->value	= ((v1 & 0xFFFF0000) >> 16) | ((v2 & 0x0000FFFF) << 16);
	}
	else if (((reg->offset & 3) == 3) && (reg->offset <= 248))
	{
		status		= 0;
		reg->status	= GSC_SUCCESS;
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset - 3,
								&v1);
		pci_read_config_dword(	dev->pci,
								(unsigned long) reg->offset + 1,
								&v2);
		reg->value	= ((v1 & 0xFF000000) >> 24) | ((v2 & 0x00FFFFFF) << 8);
	}
	else
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_OFFSET;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_read
*
*	Purpose:
*
*		Read a value from a PCI register. Register sizes may be 1, 2, 3 or 4
*		bytes long. Alignment may be 0, 1, 2 or 3 bytes. This is done
*		specifically for the CCR, though all cases are covered.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_read(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;

	switch (reg->size)
	{
		default:	status		= -EINVAL;
					reg->status	= GSC_INVALID_SIZE;
					break;

		case 1:		status	= _reg_pci_read_1(dev, reg);	break;
		case 2:		status	= _reg_pci_read_2(dev, reg);	break;
		case 3:		status	= _reg_pci_read_3(dev, reg);	break;
		case 4:		status	= _reg_pci_read_4(dev, reg);	break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_reg_pci_write
*
*	Purpose:
*
*		Write a value to a PCI register. Register sizes may be 1, 2 or 4
*		bytes long. Alignment must be per the register's size.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _reg_pci_write(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int		status;

	switch (reg->size)
	{
		default:

			status		= -EINVAL;
			reg->status	= GSC_INVALID_SIZE;
			break;

		case 1:

			if ((reg->offset + 1) >= 256)
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			pci_write_config_byte(	dev->pci,
									(unsigned long) reg->offset,
									(u8) (reg->value & 0xFF));
			break;

		case 2:

			if (((reg->offset + 2) >= 256) || (reg->offset & 1))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			pci_write_config_word(	dev->pci,
									(unsigned long) reg->offset,
									(u16) (reg->value & 0xFFFF));
			break;

		case 4:

			if (((reg->offset + 4) >= 256) || (reg->offset & 3))
			{
				status		= -EINVAL;
				reg->status	= GSC_INVALID_OFFSET;
				break;
			}

			status	= 0;
			pci_write_config_dword(	dev->pci,
									(unsigned long) reg->offset,
									reg->value & 0xFFFF);
			break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_mod_ioctl
*
*	Purpose:
*
*		Perform a read-modify-write on a register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

int reg_mod_ioctl(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;

	if ((reg->bar < -1) || (reg->bar >= (s8) ARRAY_ELEMENTS(dev->bars.bar)))
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_REGISTER;
	}
	else if (reg->bar == -1)
	{
		status	= _reg_pci_mod(dev, reg);
	}
	else if (dev->bars.bar[reg->bar].mem_spc)
	{
		status	= _reg_mem_mod(dev, reg);
	}
	else
	{
		status	= _reg_io_mod(dev, reg);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_read_ioctl
*
*	Purpose:
*
*		Read a value from a register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

int reg_read_ioctl(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;

	if ((reg->bar < -1) || (reg->bar >= (s8) ARRAY_ELEMENTS(dev->bars.bar)))
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_REGISTER;
	}
	else if (reg->bar == -1)
	{
		status	= _reg_pci_read(dev, reg);
	}
	else if (dev->bars.bar[reg->bar].mem_spc)
	{
		status	= _reg_mem_read(dev, reg);
	}
	else
	{
		status	= _reg_io_read(dev, reg);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_write_ioctl
*
*	Purpose:
*
*		Write a value to a register.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		reg	The data for the register of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

int reg_write_ioctl(dev_data_t* dev, gsc_drv_reg_t* reg)
{
	int	status;

	if ((reg->bar < -1) || (reg->bar >= (s8) ARRAY_ELEMENTS(dev->bars.bar)))
	{
		status		= -EINVAL;
		reg->status	= GSC_INVALID_REGISTER;
	}
	else if (reg->bar == -1)
	{
		status	= _reg_pci_write(dev, reg);
	}
	else if (dev->bars.bar[reg->bar].mem_spc)
	{
		status	= _reg_mem_write(dev, reg);
	}
	else
	{
		status	= _reg_io_write(dev, reg);
	}

	return(status);
}



/******************************************************************************
*
*	Function:	reg_vaddr
*
*	Purpose:
*
*		Get the virtual address of the specified register.
*
*	Arguments:
*
*		dev		The respective device of interest.
*
*		reg		The register of interest.
*
*	Returned:
*
*		VADDR_T	The register's virtual address.
*
******************************************************************************/

VADDR_T reg_vaddr(dev_data_t* dev, unsigned long reg)
{
	unsigned long	type	= GSC_REG_TYPE(reg);
	VADDR_T			vaddr;

	switch (type)
	{
		default:
		case GSC_REG_PCI:

			vaddr	= (VADDR_T) GSC_REG_OFFSET(reg);
			break;

		case GSC_REG_PLX:

			vaddr	= (VADDR_T) ((unsigned long) dev->bars.bar[0].vaddr
					+ GSC_REG_OFFSET(reg));
			break;
	}

	return(vaddr);
}



